import tensorflow as tf


class Split_Streams(tf.keras.layers.Layer):
    def __init__(self):
        super(Split_Streams, self).__init__()

    def call(self, inputs):
        x, y, z = inputs

        indice_position = tf.reshape(
            tf.cast(tf.where(tf.equal(tf.reshape(y, (-1,)), z)), tf.int32),
            (-1,),
        )

        return tf.gather(x, indice_position), indice_position



import tensorflow as tf

from ..layers.gather import Gather_Streams
from ..layers.split import Split_Streams
from ..layers.MLP import MLP
from pickle import load

from typing import Any


class TARNet(tf.keras.Model):
    """Return a tarnet sub KERAS API model."""

    def __init__(
        self,
        normalizer_layer: tf.keras.layers.Layer = None,
        n_treatments: int = 2,
        output_dim: int = 1,
        phi_layers: int = 2,
        units:int = 20,
        y_layers: int = 3,
        activation: str = "relu",
        reg_l2: float = 0.0,
        treatment_as_input: bool = False,
        scaler: Any = None,
        output_bias: float = None,
    ):
        """Initialize the layers used by the model.

        Args:
            normalizer_layer (tf.keras.layer, optional): _description_. Defaults to None.
            n_treatments (int, optional): _description_. Defaults to 2.
            output_dim (int, optional): _description_. Defaults to 1.
            phi_layers (int, optional): _description_. Defaults to 2.
            y_layers (int, optional): _description_. Defaults to 3.
            activation (str, optional): _description_. Defaults to "relu".
            reg_l2 (float, optional): _description_. Defaults to 0.0.
        """
        super(TARNet, self).__init__()
        # uniform quantile transform for treatment
        self.scaler = scaler if scaler else load(open("scaler.pkl", "rb"))

        # input normalization layer
        self.normalizer_layer = normalizer_layer
        self.phi = MLP(
            units=units,
            activation=activation,
            name="phi",
            num_layers=phi_layers,
        )

        self.splitter = Split_Streams()

        self.y_hiddens = [
            MLP(
                units=units,
                activation=activation,
                name=f"y_{k}",
                num_layers=y_layers,
            )
            for k in range(n_treatments)
        ]

        # add linear function to cover the normalized output
        self.y_outputs = [
            tf.keras.layers.Dense(
                output_dim,
                activation="sigmoid",
                bias_initializer=output_bias,
                name=f"top_{k}",
            )
            for k in range(n_treatments)
        ]

        self.n_treatments = n_treatments

        self.output_ = Gather_Streams()

    def call(self, x):

        cofeatures_input, treatment_input = x
        treatment_cat = tf.cast(treatment_input, tf.int32)

        if self.normalizer_layer:
            cofeatures_input = self.normalizer_layer(cofeatures_input)
        x_flux = self.phi(cofeatures_input)

        streams = [
            self.splitter([x_flux, treatment_cat, tf.cast(indice_treatment, tf.int32)])
            for indice_treatment in range(len(self.y_hiddens))
        ]
        # xstream is a list of tuple, containing the gathered and indice position, let's unpack them
        x_streams, indice_streams = zip(*streams)
        # tf.print(indice_streams, output_stream=sys.stderr)
        x_streams = [
            y_hidden(x_stream) for y_hidden, x_stream in zip(self.y_hiddens, x_streams)
        ]
        x_streams = [
            y_output(x_stream) for y_output, x_stream in zip(self.y_outputs, x_streams)
        ]

        return self.output_([x_streams, indice_streams])



import tensorflow as tf

from tensorflow.keras import regularizers


class MLP(tf.keras.layers.Layer):
    """Multi Layer Perceptron as a Keras layer."""

    def __init__(
        self,
        units: int = 200,
        num_layers: int = 1,
        kernel_initializer=tf.keras.initializers.HeNormal(),
        activation: str = "relu",
        name: str = "phi",
        reg_l2: float = 0,
    ):
        """Initiate the layers used by the multi layer perceptron.

        Args:
            units (int, optional): _description_. Defaults to 200.
            num_layers (int, optional): _description_. Defaults to 1.
            kernel_initializer (_type_, optional): _description_. Defaults to tf.keras.initializers.HeNormal().
            activation (str, optional): _description_. Defaults to "relu".
            name (str, optional): _description_. Defaults to "phi".
            reg_l2 (float, optional): _description_. Defaults to 0.
        """

        super(MLP, self).__init__()
        self.layers = [
            tf.keras.layers.Dense(
                units=units,
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=regularizers.l2(reg_l2),
                name=f"{name}_{k}",
                trainable=True,
            )
            for k in range(num_layers)
        ]

    def call(self, x: tf.Tensor) -> tf.Tensor:
        for layer in self.layers:
            x = layer(x)
        return x



import tensorflow as tf


class Gather_Streams(tf.keras.layers.Layer):
    def __init__(self):
        super(Gather_Streams, self).__init__()

    def call(self, inputs):
        
        x, y = inputs
        return tf.dynamic_stitch(y, x)
